# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import user_management_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2


class UserManagementServiceStub(object):
    """Service to manage users and their rights for interacting with the Ledger API
    served by a participant node.

    The authorization rules for its RPCs are specified on the ``<RpcName>Request``
    messages as boolean expressions over these two facts:
    (1) ``HasRight(r)`` denoting whether the authenticated user has right ``r`` and
    (2) ``IsAuthenticatedUser(uid)`` denoting whether ``uid`` is the empty string or equal to the id of the authenticated user.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateUser = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/CreateUser',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserResponse.FromString,
                )
        self.GetUser = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/GetUser',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserResponse.FromString,
                )
        self.DeleteUser = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/DeleteUser',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserResponse.FromString,
                )
        self.ListUsers = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/ListUsers',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersResponse.FromString,
                )
        self.GrantUserRights = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/GrantUserRights',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsResponse.FromString,
                )
        self.RevokeUserRights = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/RevokeUserRights',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsResponse.FromString,
                )
        self.ListUserRights = channel.unary_unary(
                '/com.daml.ledger.api.v1.admin.UserManagementService/ListUserRights',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsResponse.FromString,
                )


class UserManagementServiceServicer(object):
    """Service to manage users and their rights for interacting with the Ledger API
    served by a participant node.

    The authorization rules for its RPCs are specified on the ``<RpcName>Request``
    messages as boolean expressions over these two facts:
    (1) ``HasRight(r)`` denoting whether the authenticated user has right ``r`` and
    (2) ``IsAuthenticatedUser(uid)`` denoting whether ``uid`` is the empty string or equal to the id of the authenticated user.

    """

    def CreateUser(self, request, context):
        """Create a new user.
        Errors:
        - ``ALREADY_EXISTS``: if the user already exists
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUser(self, request, context):
        """Get the user data of a specific user or the authenticated user.
        Errors:
        - ``NOT_FOUND``: if the user doesn't exist
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteUser(self, request, context):
        """Delete an existing user and all its rights.
        Errors:
        - ``NOT_FOUND``: if the user doesn't exist
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListUsers(self, request, context):
        """List all existing users.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GrantUserRights(self, request, context):
        """Grant rights to a user.
        Errors:
        - ``NOT_FOUND``: if the user doesn't exist
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RevokeUserRights(self, request, context):
        """Revoke rights from a user.
        Errors:
        - ``NOT_FOUND``: if the user doesn't exist
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListUserRights(self, request, context):
        """List the set of all rights granted to a user.
        Errors:
        - ``NOT_FOUND``: if the user doesn't exist
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UserManagementServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateUser': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateUser,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserResponse.SerializeToString,
            ),
            'GetUser': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUser,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserResponse.SerializeToString,
            ),
            'DeleteUser': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteUser,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserResponse.SerializeToString,
            ),
            'ListUsers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUsers,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersResponse.SerializeToString,
            ),
            'GrantUserRights': grpc.unary_unary_rpc_method_handler(
                    servicer.GrantUserRights,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsResponse.SerializeToString,
            ),
            'RevokeUserRights': grpc.unary_unary_rpc_method_handler(
                    servicer.RevokeUserRights,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsResponse.SerializeToString,
            ),
            'ListUserRights': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUserRights,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.admin.UserManagementService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UserManagementService(object):
    """Service to manage users and their rights for interacting with the Ledger API
    served by a participant node.

    The authorization rules for its RPCs are specified on the ``<RpcName>Request``
    messages as boolean expressions over these two facts:
    (1) ``HasRight(r)`` denoting whether the authenticated user has right ``r`` and
    (2) ``IsAuthenticatedUser(uid)`` denoting whether ``uid`` is the empty string or equal to the id of the authenticated user.

    """

    @staticmethod
    def CreateUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/CreateUser',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.CreateUserResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/GetUser',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GetUserResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/DeleteUser',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.DeleteUserResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListUsers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/ListUsers',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUsersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GrantUserRights(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/GrantUserRights',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.GrantUserRightsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RevokeUserRights(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/RevokeUserRights',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.RevokeUserRightsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListUserRights(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.admin.UserManagementService/ListUserRights',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_user__management__service__pb2.ListUserRightsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Collection, Iterator, TypeVar
import warnings

from ..damlast.lookup import matching_normalizations, normalize, validate_template
from ..protocols.events import (
    BaseEvent,
    ContractArchiveEvent,
    ContractCreateEvent,
    ContractExercisedEvent,
    InitEvent,
    OffsetEvent,
    PackagesAddedEvent,
    ReadyEvent,
    TransactionEndEvent,
    TransactionStartEvent,
)

__all__ = ["EventKey", "_template_reverse_globs"]

T = TypeVar("T")


def create_dispatch(
    on_init: "Callable[[InitEvent], T]",
    on_ready: "Callable[[ReadyEvent], T]",
    on_offset: "Callable[[OffsetEvent], T]",
    on_transaction_start: "Callable[[TransactionStartEvent], T]",
    on_transaction_end: "Callable[[TransactionEndEvent], T]",
    on_contract_created: "Callable[[ContractCreateEvent], T]",
    on_contract_exercised: "Callable[[ContractExercisedEvent], T]",
    on_contract_archived: "Callable[[ContractArchiveEvent], T]",
    on_packages_added: "Callable[[PackagesAddedEvent], T]",
) -> "Callable[[BaseEvent], T]":
    def handle(event: "BaseEvent") -> "T":
        if isinstance(event, ContractCreateEvent):
            return on_contract_created(event)
        elif isinstance(event, ContractExercisedEvent):
            return on_contract_exercised(event)
        elif isinstance(event, ContractArchiveEvent):
            return on_contract_archived(event)
        elif isinstance(event, TransactionStartEvent):
            return on_transaction_start(event)
        elif isinstance(event, TransactionEndEvent):
            return on_transaction_end(event)
        elif isinstance(event, ReadyEvent):
            return on_ready(event)
        elif isinstance(event, InitEvent):
            return on_init(event)
        elif isinstance(event, OffsetEvent):
            return on_offset(event)
        elif isinstance(event, PackagesAddedEvent):
            return on_packages_added(event)
        else:
            raise ValueError(f"unknown subclass of BaseEvent: {event!r}")

    return handle


def _template_reverse_globs(primary_only: bool, package_id: str, type_name: str) -> "Iterator[str]":
    """
    Return an iterator over strings that glob to a specified type.

    This symbol is meant only for supporting the dazl v7 and is not a public API!
    """
    # support deprecated type identifiers for usages of this old API to preserve backwards
    # compatibility
    use_deprecated_form = False
    if type_name != "*" and ":" not in type_name:
        m, delim, e = type_name.rpartition(".")
        if delim:
            use_deprecated_form = True
            type_name = f"{m}:{e}"
    t = f"{package_id}:{type_name}"

    if primary_only:
        yield normalize(t)
    else:
        for form in matching_normalizations(t):
            yield form
            if use_deprecated_form and len(form.split(":")) == 3:
                s, _, e = form.rpartition(":")
                yield f"{s}.{e}"


class EventKey:
    from_event = create_dispatch(
        on_init=lambda _: EventKey.init(),
        on_ready=lambda _: EventKey.ready(),
        on_offset=lambda _: EventKey.offset(),
        on_transaction_start=lambda _: EventKey.transaction_start(),
        on_transaction_end=lambda _: EventKey.transaction_end(),
        on_contract_created=lambda event: EventKey.contract_created(False, event.cid.value_type),
        on_contract_exercised=lambda event: EventKey.contract_exercised(
            False, event.cid.value_type, event.choice
        ),
        on_contract_archived=lambda event: EventKey.contract_archived(False, event.cid.value_type),
        on_packages_added=lambda event: EventKey.packages_added(
            initial=event.initial, changed=not event.initial
        ),
    )

    @staticmethod
    def init() -> Collection[str]:
        """
        Return the names of events that get raised in response to an :class:`InitEvent`. This is
        currently only ``'init'``.
        """
        return ("init",)

    @staticmethod
    def ready() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ReadyEvent`. This is
        currently only ``'ready'``.
        """
        return ("ready",)

    @staticmethod
    def offset() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`OffsetEvent`. This is
        currently only ``'offset'``.
        """
        return ("offset",)

    @staticmethod
    def transaction_start() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionStartEvent`.
        This is currently only ``'transaction-start'``.
        """
        return ("transaction-start",)

    @staticmethod
    def transaction_end() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionEndEvent`.
        This is currently only ``'transaction-end'``.
        """
        return ("transaction-end",)

    @staticmethod
    def contract_created(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, "create", template)

    @staticmethod
    def contract_exercised(primary_only: bool, template: Any, choice: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractExercisedEvent`
        of the specified choice.
        """
        return [
            f"{key}/{choice}" for key in EventKey._contract(primary_only, "exercised", template)
        ]

    @staticmethod
    def contract_archived(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, "archive", template)

    @staticmethod
    def packages_added(initial: bool, changed: bool) -> "Collection[str]":
        keys = []
        if initial:
            keys.append("packages-added/initial")
        if changed:
            keys.append("packages-added/changed")
        return tuple(keys)

    @staticmethod
    def _contract(primary_only: bool, prefix: str, template: "Any") -> Collection[str]:
        m, t = validate_template(template)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return tuple(f"{prefix}/{g}" for g in _template_reverse_globs(primary_only, m, t))



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/completion.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.rpc import status_pb2 as google_dot_rpc_dot_status__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'com/daml/ledger/api/v1/completion.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\x1egoogle/protobuf/duration.proto\x1a\x17google/rpc/status.proto\"\x99\x03\n\nCompletion\x12\x1d\n\ncommand_id\x18\x01 \x01(\tR\tcommandId\x12*\n\x06status\x18\x02 \x01(\x0b\x32\x12.google.rpc.StatusR\x06status\x12%\n\x0etransaction_id\x18\x03 \x01(\tR\rtransactionId\x12%\n\x0e\x61pplication_id\x18\x04 \x01(\tR\rapplicationId\x12\x15\n\x06\x61\x63t_as\x18\x05 \x03(\tR\x05\x61\x63tAs\x12#\n\rsubmission_id\x18\x06 \x01(\tR\x0csubmissionId\x12\x33\n\x14\x64\x65\x64uplication_offset\x18\x08 \x01(\tH\x00R\x13\x64\x65\x64uplicationOffset\x12R\n\x16\x64\x65\x64uplication_duration\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x15\x64\x65\x64uplicationDurationB\x16\n\x14\x64\x65\x64uplication_periodJ\x04\x08\x07\x10\x08R\x0fsubmission_rankB\x8e\x01\n\x16\x63om.daml.ledger.api.v1B\x14\x43ompletionOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_COMPLETION = DESCRIPTOR.message_types_by_name['Completion']
Completion = _reflection.GeneratedProtocolMessageType('Completion', (_message.Message,), {
  'DESCRIPTOR' : _COMPLETION,
  '__module__' : 'com.daml.ledger.api.v1.completion_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Completion)
  })
_sym_db.RegisterMessage(Completion)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\024CompletionOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _COMPLETION._serialized_start=125
  _COMPLETION._serialized_end=534
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Conversion methods to Ledger API Protobuf-generated types from dazl/Pythonic types.
"""
from typing import TYPE_CHECKING, Any, Union
import warnings

from ..._gen.com.daml.ledger.api import v1 as lapipb
from ...damlast.daml_lf_1 import TypeConName
from ...damlast.util import module_local_name, module_name, package_ref
from ...ledger.grpc.codec_aio import Codec
from ...prim import ContractId, timedelta_to_duration
from ...values.protobuf import ProtobufEncoder, set_value
from ..serializers import AbstractSerializer

if TYPE_CHECKING:
    from ...client.commands import CommandPayload
    from ...model.types import TypeReference
    from ...model.writing import CommandPayload

__all__ = ["as_identifier", "ProtobufSerializer"]


def as_identifier(tref: "Union[TypeReference, TypeConName]") -> "lapipb.Identifier":
    from ...model.types import TypeReference

    warnings.warn("Use Codec.encode_identifier instead.", DeprecationWarning, stacklevel=2)

    if isinstance(tref, TypeReference):
        tref = tref.con

    return Codec.encode_identifier(tref)


class ProtobufSerializer(AbstractSerializer):

    mapper = ProtobufEncoder()

    ################################################################################################
    # COMMAND serializers
    ################################################################################################

    def serialize_command_request(
        self, command_payload: "CommandPayload"
    ) -> "lapipb.SubmitAndWaitRequest":
        commands = [self.serialize_command(command) for command in command_payload.commands]
        if command_payload.deduplication_time is not None:
            return lapipb.SubmitAndWaitRequest(
                commands=lapipb.Commands(
                    ledger_id=command_payload.ledger_id,
                    workflow_id=command_payload.workflow_id,
                    application_id=command_payload.application_id,
                    command_id=command_payload.command_id,
                    party=command_payload.party,
                    commands=commands,
                    deduplication_time=timedelta_to_duration(command_payload.deduplication_time),
                )
            )
        else:
            return lapipb.SubmitAndWaitRequest(
                commands=lapipb.Commands(
                    ledger_id=command_payload.ledger_id,
                    workflow_id=command_payload.workflow_id,
                    application_id=command_payload.application_id,
                    command_id=command_payload.command_id,
                    party=command_payload.party,
                    commands=commands,
                )
            )

    def serialize_create_command(
        self, name: "TypeConName", template_args: "Any"
    ) -> "lapipb.Command":
        create_ctor, create_value = template_args
        if create_ctor != "record":
            raise ValueError("Template values must resemble records")

        cmd = lapipb.CreateCommand()
        _set_template(cmd.template_id, name)
        cmd.create_arguments.MergeFrom(create_value)
        return lapipb.Command(create=cmd)

    def serialize_exercise_command(
        self, contract_id: "ContractId", choice_name: str, choice_args: "Any"
    ) -> lapipb.Command:
        type_ref = contract_id.value_type
        ctor, value = choice_args

        cmd = lapipb.ExerciseCommand()
        _set_template(cmd.template_id, type_ref)
        cmd.contract_id = contract_id.value
        cmd.choice = choice_name
        set_value(cmd.choice_argument, ctor, value)
        return lapipb.Command(exercise=cmd)

    def serialize_exercise_by_key_command(
        self,
        template_name: "TypeConName",
        key_arguments: Any,
        choice_name: str,
        choice_arguments: Any,
    ) -> lapipb.Command:
        key_ctor, key_value = key_arguments
        choice_ctor, choice_value = choice_arguments

        cmd = lapipb.ExerciseByKeyCommand()
        _set_template(cmd.template_id, template_name)
        set_value(cmd.contract_key, key_ctor, key_value)
        cmd.choice = choice_name
        set_value(cmd.choice_argument, choice_ctor, choice_value)
        return lapipb.Command(exerciseByKey=cmd)

    def serialize_create_and_exercise_command(
        self,
        template_name: "TypeConName",
        create_arguments: "Any",
        choice_name: str,
        choice_arguments: Any,
    ) -> lapipb.Command:
        create_ctor, create_value = create_arguments
        if create_ctor != "record":
            raise ValueError("Template values must resemble records")
        choice_ctor, choice_value = choice_arguments

        cmd = lapipb.CreateAndExerciseCommand()
        _set_template(cmd.template_id, template_name)
        cmd.create_arguments.MergeFrom(create_value)
        cmd.choice = choice_name
        set_value(cmd.choice_argument, choice_ctor, choice_value)
        return lapipb.Command(createAndExercise=cmd)


def _set_template(message: "lapipb.Identifier", name: "TypeConName") -> None:
    message.package_id = package_ref(name)
    message.module_name = str(module_name(name))
    message.entity_name = module_local_name(name)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ..protoc import protoc_plugin
from .plugin import run_all_plugins

__all__ = ["main"]

main = protoc_plugin(run_all_plugins)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .command_completion_service_pb2 import CompletionEndRequest, CompletionEndResponse, CompletionStreamRequest, CompletionStreamResponse

__all__ = [
    "CommandCompletionServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class CommandCompletionServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _CommandCompletionServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _CommandCompletionServiceAsyncStub: ...  # type: ignore
    def CompletionStream(self, __1: CompletionStreamRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[_grpc.CallIterator[CompletionStreamResponse], _grpc_aio.UnaryStreamCall[_typing.Any, CompletionStreamResponse]]: ...
    def CompletionEnd(self, __1: CompletionEndRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[CompletionEndResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, CompletionEndResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _CommandCompletionServiceBlockingStub(CommandCompletionServiceStub):
    def CompletionStream(self, __1: CompletionStreamRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc.CallIterator[CompletionStreamResponse]: ...
    def CompletionEnd(self, __1: CompletionEndRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> CompletionEndResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _CommandCompletionServiceAsyncStub(CommandCompletionServiceStub):
    def CompletionStream(self, __1: CompletionStreamRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryStreamCall[_typing.Any, CompletionStreamResponse]: ...
    def CompletionEnd(self, __1: CompletionEndRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, CompletionEndResponse]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Internal process communications that apply globally. These include signal handlers.
"""

from collections import defaultdict
from threading import RLock
from typing import Any, Callable

from .. import LOG

_HANDLER_LOCK = RLock()
_HANDLERS = defaultdict(list)


def add_handler(key: str, cb: Callable[[Any], None]):
    if key is None:
        raise ValueError("key is required")
    if cb is None:
        raise ValueError("callable is required")

    with _HANDLER_LOCK:
        _HANDLERS[key].append(cb)


def remove_handler(key: str, cb: Callable[[Any], None]):
    if key is None:
        raise ValueError("key is required")
    if cb is None:
        raise ValueError("callable is required")

    with _HANDLER_LOCK:
        _HANDLERS[key].remove(cb)


def invoke_handlers(key: str, arg: Any):
    with _HANDLER_LOCK:
        if key in _HANDLERS:
            handlers = list(_HANDLERS[key])
        else:
            return

    for handler in handlers:
        try:
            handler(arg)
        except:
            LOG.exception()



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Utilities for working with the built-in `logging` module.

This module will be removed soon!
"""
import logging
import warnings

from .._logging import VERBOSE, configure

__all__ = ["VERBOSE", "setup_default_logger"]


def setup_default_logger(level=logging.INFO):
    """
    Sets up a default logger with sensible defaults.

    This function is deprecated, and there is no replacement; you should instead configure your own
    logging output as suitable for your needs (``logging.basicConfig()`` may suffice).
    """
    warnings.warn("dazl.setup_default_logger is deprecated", DeprecationWarning, stacklevel=2)
    configure(level)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Sequence as Seq
from os import PathLike
import sys
from typing import (
    AsyncContextManager,
    BinaryIO,
    Callable,
    Collection,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

from ..ledger.aio import Connection
from ..prim import Party

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal
__all__ = ["connect_with_new_party"]
NameGenFn = Callable[[int], Optional[str]]

@overload
def connect_with_new_party(
    *,
    party_count: Literal[1] = 1,
    url: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    dar: Union[str, bytes, PathLike, BinaryIO] = None,
    identifier_hint: Union[None, str, NameGenFn] = None,
    display_name: Union[None, str, NameGenFn] = None
) -> AsyncContextManager[ConnectionWithParty]: ...
@overload
def connect_with_new_party(
    *,
    party_count: Literal[2],
    url: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    dar: Union[str, bytes, PathLike, BinaryIO] = None,
    identifier_hint: Union[None, str, NameGenFn] = None,
    display_name: Union[None, str, NameGenFn] = None
) -> AsyncContextManager[Tuple[ConnectionWithParty, ConnectionWithParty]]: ...
@overload
def connect_with_new_party(
    *,
    party_count: Literal[3],
    url: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    dar: Union[str, bytes, PathLike, BinaryIO] = None,
    identifier_hint: Union[None, str, NameGenFn] = None,
    display_name: Union[None, str, NameGenFn] = None
) -> AsyncContextManager[Tuple[ConnectionWithParty, ConnectionWithParty, ConnectionWithParty]]: ...
@overload
def connect_with_new_party(
    *,
    party_count: Literal[4],
    url: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    dar: Union[str, bytes, PathLike, BinaryIO] = None,
    identifier_hint: Union[None, str, NameGenFn] = None,
    display_name: Union[None, str, NameGenFn] = None
) -> AsyncContextManager[
    Tuple[ConnectionWithParty, ConnectionWithParty, ConnectionWithParty, ConnectionWithParty]
]: ...
@overload
def connect_with_new_party(
    *,
    party_count: int,
    url: Optional[str] = None,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = False,
    ledger_id: Optional[str] = None,
    dar: Union[str, bytes, PathLike, BinaryIO] = None,
    identifier_hint: Union[None, str, NameGenFn] = None,
    display_name: Union[None, str, NameGenFn] = None
) -> AsyncContextManager[Sequence[ConnectionWithParty]]: ...

class ConnectionWithParty(Seq["ConnectionWithParty"]):
    def __init__(self, connection: Connection, party: Party): ...
    @property
    def connection(self) -> Connection: ...
    @property
    def party(self) -> Party: ...
    def __getitem__(self, item): ...
    def __len__(self) -> Literal[1]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
from typing import Iterable

__all__ = ["all_decl"]


def all_decl(__names: "Iterable[str]") -> str:
    if __names:
        with StringIO() as buf:
            buf.write("__all__ = [\n")
            for name in __names:
                buf.write(f'    "{name}",\n')
            buf.write("]\n")

            return buf.getvalue()
    else:
        return "__all__ = []  # type: ignore\n"



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

from .time_service_pb2 import GetTimeRequest, GetTimeResponse, SetTimeRequest
from .time_service_pb2_grpc import TimeServiceStub

__all__ = [
    "GetTimeRequest",
    "GetTimeResponse",
    "SetTimeRequest",
    "TimeServiceStub",
]



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import MessageMap, RepeatedCompositeFieldContainer
from google.protobuf.message import Message as _Message

from .value_pb2 import Identifier

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "TransactionFilter",
    "Filters",
    "InclusiveFilters",
]


class TransactionFilter(_Message):
    @property
    def filters_by_party(self) -> MessageMap[_builtins.str, Filters]: ...
    def __init__(self, *, filters_by_party: _typing.Optional[_typing.Mapping[_builtins.str, Filters]] = ...): ...
    def HasField(self, field_name: _L["filters_by_party"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["filters_by_party"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class Filters(_Message):
    @property
    def inclusive(self) -> InclusiveFilters: ...
    def __init__(self, *, inclusive: _typing.Optional[InclusiveFilters] = ...): ...
    def HasField(self, field_name: _L["inclusive"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["inclusive"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class InclusiveFilters(_Message):
    @property
    def template_ids(self) -> RepeatedCompositeFieldContainer[Identifier]: ...
    def __init__(self, *, template_ids: _typing.Optional[_typing.Iterable[Identifier]] = ...): ...
    def HasField(self, field_name: _L["template_ids"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["template_ids"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
The core scheduling of logic, managing the tricky interaction between the man asyncio event loop and
background threads.
"""
from asyncio import CancelledError, Future, InvalidStateError, gather, get_event_loop, wait_for
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
import signal
from typing import List
import warnings

from ..prim import TimeDeltaLike, to_timedelta
from ..util.asyncio_util import execute_in_loop, safe_create_future
from ._base import RunLevel

__all__ = ["Invoker", "DEFAULT_TIMEOUT"]

DEFAULT_TIMEOUT = timedelta(seconds=30)


class Invoker:
    """
    A generally thread-safe invoker that aids in coordination between event-loop driven events and
    background thread events.

    This serves a similar purpose to Akka's ExecutionContext in Scala.
    """

    def __init__(self) -> None:
        self.level = RunLevel.RUN_FOREVER
        self.loop = None
        self.executor = None
        self._futures = []  # type: List[Future]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.shutdown(exception=exc_val)

    def _unhook_future(self, fut: Future):
        try:
            self._futures.remove(fut)
        except ValueError:
            pass

    def create_future(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            f = safe_create_future()
        f.add_done_callback(self._unhook_future)
        self._futures.append(f)
        return f

    async def shutdown(self, timeout=DEFAULT_TIMEOUT, exception=None) -> None:
        """
        Stop the event loop and executor. Outstanding futures will be terminated after the
        specified timeout, or if an error is provided, that error will be used to terminate all
        futures IMMEDIATELY.

        :param timeout:
            The maximum amount of time before outstanding Futures are terminated ungracefully.
        :param exception:
            If provided, is used to trigger all outstanding futures as failures.
        """
        t = to_timedelta(timeout)
        if exception is not None:
            for fut in self._futures:
                if not fut.done():
                    fut.set_exception(exception)
        else:
            if self._futures:
                if t.total_seconds() > 0:
                    try:
                        await wait_for(gather(*self._futures), timeout=t.total_seconds())
                    except CancelledError:
                        for fut in self._futures:
                            if not fut.done():
                                fut.cancel()
                else:
                    for fut in self._futures:
                        if not fut.done():
                            fut.cancel()

    def set_context_as_current(self) -> None:
        """
        Adopt the current event loop as the loop for this :class:`Invoker`, and additionally define
        a default executor if one has not yet been set.
        """
        self.loop = get_event_loop()
        if self.executor is None:
            self.executor = ThreadPoolExecutor()

    def run_in_loop(self, func, timeout: TimeDeltaLike = 30.0):
        """
        Schedule a normal function or coroutine function to be run on the event loop, and block
        until the function has returned.
        """
        # TODO: the awful awful witchcraft required to remove these checks
        if self.loop is None:
            raise InvalidStateError("loop must be set before calling these methods")
        return execute_in_loop(self.loop, func, timeout=timeout)

    def run_in_executor(self, func):
        """
        Schedule a normal function to be run on a background thread, and yield until the function
        has returned.
        """
        # TODO: the awful awful witchcraft required to remove these checks
        if self.loop is None or self.executor is None:
            raise InvalidStateError("loop must be set before calling these methods")

        return self.loop.run_in_executor(self.executor, func)

    def install_signal_handlers(self) -> None:
        try:
            if self.loop is not None:
                self.loop.add_signal_handler(signal.SIGINT, self.handle_sigint)
                self.loop.add_signal_handler(signal.SIGQUIT, self.handle_sigquit)
            else:
                signal.signal(signal.SIGINT, lambda *_: self.handle_sigint())
                signal.signal(signal.SIGQUIT, lambda *_: self.handle_sigquit())
        except (NotImplementedError, AttributeError, ValueError):
            # SIGINT and SIGQUIT are not supported on Windows.
            pass

    def handle_sigint(self) -> None:
        self.level = RunLevel.TERMINATE_GRACEFULLY

    def handle_sigquit(self) -> None:
        self.level = RunLevel.TERMINATE_IMMEDIATELY



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import reset_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_testing_dot_reset__service__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class ResetServiceStub(object):
    """Service to reset the ledger state. The goal here is to be able to reset the state in a way
    that's much faster compared to restarting the whole ledger application (be it a sandbox
    or the real ledger server).

    Note that *all* state present in the ledger implementation will be reset, most importantly
    including the ledger ID. This means that clients will have to re-fetch the ledger ID
    from the identity service after hitting this endpoint.

    The semantics are as follows:

    * When the reset service returns the reset is initiated, but not completed;
    * While the reset is performed, the ledger will not accept new requests. In fact we guarantee
    that ledger stops accepting new requests by the time the response to Reset is delivered;
    * In-flight requests might be aborted, we make no guarantees on when or how quickly this
    happens;
    * The ledger might be unavailable for a period of time before the reset is complete.

    Given the above, the recommended mode of operation for clients of the reset endpoint is to
    call it, then call the ledger identity endpoint in a retry loop that will tolerate a brief
    window when the ledger is down, and resume operation as soon as the new ledger ID is delivered.

    Note that this service will be available on the sandbox and might be available in some other testing
    environments, but will *never* be available in production.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Reset = channel.unary_unary(
                '/com.daml.ledger.api.v1.testing.ResetService/Reset',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_testing_dot_reset__service__pb2.ResetRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class ResetServiceServicer(object):
    """Service to reset the ledger state. The goal here is to be able to reset the state in a way
    that's much faster compared to restarting the whole ledger application (be it a sandbox
    or the real ledger server).

    Note that *all* state present in the ledger implementation will be reset, most importantly
    including the ledger ID. This means that clients will have to re-fetch the ledger ID
    from the identity service after hitting this endpoint.

    The semantics are as follows:

    * When the reset service returns the reset is initiated, but not completed;
    * While the reset is performed, the ledger will not accept new requests. In fact we guarantee
    that ledger stops accepting new requests by the time the response to Reset is delivered;
    * In-flight requests might be aborted, we make no guarantees on when or how quickly this
    happens;
    * The ledger might be unavailable for a period of time before the reset is complete.

    Given the above, the recommended mode of operation for clients of the reset endpoint is to
    call it, then call the ledger identity endpoint in a retry loop that will tolerate a brief
    window when the ledger is down, and resume operation as soon as the new ledger ID is delivered.

    Note that this service will be available on the sandbox and might be available in some other testing
    environments, but will *never* be available in production.
    """

    def Reset(self, request, context):
        """Resets the ledger state. Note that loaded DARs won't be removed -- this only rolls back the
        ledger to genesis.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ResetServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Reset': grpc.unary_unary_rpc_method_handler(
                    servicer.Reset,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_testing_dot_reset__service__pb2.ResetRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.testing.ResetService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ResetService(object):
    """Service to reset the ledger state. The goal here is to be able to reset the state in a way
    that's much faster compared to restarting the whole ledger application (be it a sandbox
    or the real ledger server).

    Note that *all* state present in the ledger implementation will be reset, most importantly
    including the ledger ID. This means that clients will have to re-fetch the ledger ID
    from the identity service after hitting this endpoint.

    The semantics are as follows:

    * When the reset service returns the reset is initiated, but not completed;
    * While the reset is performed, the ledger will not accept new requests. In fact we guarantee
    that ledger stops accepting new requests by the time the response to Reset is delivered;
    * In-flight requests might be aborted, we make no guarantees on when or how quickly this
    happens;
    * The ledger might be unavailable for a period of time before the reset is complete.

    Given the above, the recommended mode of operation for clients of the reset endpoint is to
    call it, then call the ledger identity endpoint in a retry loop that will tolerate a brief
    window when the ledger is down, and resume operation as soon as the new ledger ID is delivered.

    Note that this service will be available on the sandbox and might be available in some other testing
    environments, but will *never* be available in production.
    """

    @staticmethod
    def Reset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.testing.ResetService/Reset',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_testing_dot_reset__service__pb2.ResetRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "User",
    "Right",
    "CreateUserRequest",
    "CreateUserResponse",
    "GetUserRequest",
    "GetUserResponse",
    "DeleteUserRequest",
    "DeleteUserResponse",
    "ListUsersRequest",
    "ListUsersResponse",
    "GrantUserRightsRequest",
    "GrantUserRightsResponse",
    "RevokeUserRightsRequest",
    "RevokeUserRightsResponse",
    "ListUserRightsRequest",
    "ListUserRightsResponse",
]


class User(_Message):
    id: _builtins.str
    primary_party: _builtins.str
    def __init__(self, *, id: _typing.Optional[_builtins.str] = ..., primary_party: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["id", "primary_party"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["id", "primary_party"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class Right(_Message):
    
    class ParticipantAdmin(_Message):
        def __init__(self): ...
        def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
        def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
        def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...
    
    class CanActAs(_Message):
        party: _builtins.str
        def __init__(self, *, party: _typing.Optional[_builtins.str] = ...): ...
        def HasField(self, field_name: _L["party"]) -> _builtins.bool: ...
        def ClearField(self, field_name: _L["party"]) -> None: ...
        def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...
    
    class CanReadAs(_Message):
        party: _builtins.str
        def __init__(self, *, party: _typing.Optional[_builtins.str] = ...): ...
        def HasField(self, field_name: _L["party"]) -> _builtins.bool: ...
        def ClearField(self, field_name: _L["party"]) -> None: ...
        def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...
    @property
    def participant_admin(self) -> Right.ParticipantAdmin: ...
    @property
    def can_act_as(self) -> Right.CanActAs: ...
    @property
    def can_read_as(self) -> Right.CanReadAs: ...
    @_typing.overload
    def __init__(self): ...
    @_typing.overload
    def __init__(self, *, participant_admin: Right.ParticipantAdmin = ...): ...
    @_typing.overload
    def __init__(self, *, can_act_as: Right.CanActAs = ...): ...
    @_typing.overload
    def __init__(self, *, can_read_as: Right.CanReadAs = ...): ...
    def HasField(self, field_name: _L["kind", "participant_admin", "can_act_as", "can_read_as"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["kind", "participant_admin", "can_act_as", "can_read_as"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["kind"]) -> _L[None, "participant_admin", "can_act_as", "can_read_as"]: ...

class CreateUserRequest(_Message):
    @property
    def user(self) -> User: ...
    @property
    def rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, user: _typing.Optional[User] = ..., rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["user", "rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user", "rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CreateUserResponse(_Message):
    @property
    def user(self) -> User: ...
    def __init__(self, *, user: _typing.Optional[User] = ...): ...
    def HasField(self, field_name: _L["user"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetUserRequest(_Message):
    user_id: _builtins.str
    def __init__(self, *, user_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["user_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetUserResponse(_Message):
    @property
    def user(self) -> User: ...
    def __init__(self, *, user: _typing.Optional[User] = ...): ...
    def HasField(self, field_name: _L["user"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class DeleteUserRequest(_Message):
    user_id: _builtins.str
    def __init__(self, *, user_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["user_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class DeleteUserResponse(_Message):
    def __init__(self): ...
    def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListUsersRequest(_Message):
    page_token: _builtins.str
    page_size: _builtins.int
    def __init__(self, *, page_token: _typing.Optional[_builtins.str] = ..., page_size: _typing.Optional[_builtins.int] = ...): ...
    def HasField(self, field_name: _L["page_token", "page_size"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["page_token", "page_size"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListUsersResponse(_Message):
    @property
    def users(self) -> RepeatedCompositeFieldContainer[User]: ...
    next_page_token: _builtins.str
    def __init__(self, *, users: _typing.Optional[_typing.Iterable[User]] = ..., next_page_token: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["users", "next_page_token"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["users", "next_page_token"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GrantUserRightsRequest(_Message):
    user_id: _builtins.str
    @property
    def rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, user_id: _typing.Optional[_builtins.str] = ..., rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["user_id", "rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user_id", "rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GrantUserRightsResponse(_Message):
    @property
    def newly_granted_rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, newly_granted_rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["newly_granted_rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["newly_granted_rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class RevokeUserRightsRequest(_Message):
    user_id: _builtins.str
    @property
    def rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, user_id: _typing.Optional[_builtins.str] = ..., rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["user_id", "rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user_id", "rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class RevokeUserRightsResponse(_Message):
    @property
    def newly_revoked_rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, newly_revoked_rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["newly_revoked_rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["newly_revoked_rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListUserRightsRequest(_Message):
    user_id: _builtins.str
    def __init__(self, *, user_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["user_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["user_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListUserRightsResponse(_Message):
    @property
    def rights(self) -> RepeatedCompositeFieldContainer[Right]: ...
    def __init__(self, *, rights: _typing.Optional[_typing.Iterable[Right]] = ...): ...
    def HasField(self, field_name: _L["rights"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["rights"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "LedgerOffset",
]


class LedgerOffset(_Message):
    absolute: _builtins.str
    @property
    def boundary(self) -> _L[0, 1]: ...
    @_typing.overload
    def __init__(self): ...
    @_typing.overload
    def __init__(self, *, absolute: _builtins.str = ...): ...
    @_typing.overload
    def __init__(self, *, boundary: _L['LEDGER_BEGIN', 0, 'LEDGER_END', 1] = ...): ...
    def HasField(self, field_name: _L["value", "absolute", "boundary"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["value", "absolute", "boundary"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["value"]) -> _L[None, "absolute", "boundary"]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import Future, get_event_loop, sleep
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import Awaitable, Collection, Dict, List, Optional, Tuple, Union, cast
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.lookup import parse_type_con_name
from ..damlast.protocols import SymbolLookup
from ..damlast.util import package_ref
from ..prim import ContractData, ContractId
from ..protocols.events import ContractArchiveEvent, ContractCreateEvent
from ..query import ContractMatch, is_match
from ..scheduler import Invoker
from ..util.asyncio_util import await_then
from .errors import UnknownTemplateWarning

__all__ = [
    "ContractsState",
    "ContractsHistoricalState",
    "ContractContextualData",
    "ContractContextualDataCollection",
    "ActiveContractSet",
    "PendingQuery",
]


ContractsState = Dict[ContractId, ContractData]
ContractsHistoricalState = Dict[ContractId, Tuple[ContractData, bool]]


class ContractContextualDataCollection(tuple):
    def __getitem__(self, index):
        if index is None:
            raise ValueError("the index cannot be None")
        elif isinstance(index, int):
            return tuple.__getitem__(self, index)
        elif isinstance(index, str):
            for cxd in self:
                if cxd.cid.contract_id == index:
                    return cxd
            raise KeyError(index)
        elif isinstance(index, ContractId):
            for cxd in self:
                if cxd.cid == index:
                    return cxd
            raise KeyError(index)
        else:
            raise TypeError("cannot index into a ContractContextualDataCollection with {index!r}")


@dataclass(frozen=True)
class ContractContextualData:
    cid: ContractId
    cdata: Optional[ContractData]
    effective_at: Optional[datetime]
    archived_at: Optional[datetime]
    active: bool


class ActiveContractSet:
    def __init__(self, invoker: Optional[Invoker], lookup: SymbolLookup):
        self.invoker = invoker
        self.lookup = lookup
        self._tcdata = defaultdict(
            TemplateContractData
        )  # type: Dict[TypeConName, TemplateContractData]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._tcdata[event.cid.value_type].handle_create(event)

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        self._tcdata[event.cid.value_type].handle_archive(event)

    def get(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        """
        Return information for the associated :class:`ContractId`.

        :param cid: Either a ContractId or a string that represents a ContractId.
        :return: Contract information for the specified contract ID.
        """
        tcd = self._tcdata.get(cid.value_type) if isinstance(cid, ContractId) else None
        if tcd is not None:
            return tcd.get(cid)
        else:
            # untyped ContractId lookup, so scan across all templates
            for tcd in self._tcdata.values():
                cxd = tcd.get(cid)
                if cxd is not None:
                    return cxd
            return None

    def read_active(self, template_name: str, match: ContractMatch = None) -> ContractsState:
        return {cxd.cid: cxd.cdata for cxd in self.read_full(template_name, match)}

    def read_full(
        self, template_name: str, match: ContractMatch = None, include_archived: bool = False
    ) -> ContractContextualDataCollection:
        unfiltered = self._get_template_state(template_name)
        all_items = [
            item for tcd in unfiltered.values() for item in tcd.subset(match, include_archived)
        ]
        return ContractContextualDataCollection(all_items)

    async def read_async(
        self, template_name: str, match: ContractMatch = None, min_count: int = 1
    ) -> ContractsState:
        from .. import LOG

        # Fetch matching names; we may need to wait for a package to be fetched in order to do this.
        # Note that this loop may never terminate; in this case, this behavior is actually
        # acceptable because it is what enables us to wait indefinitely for the creation of a
        # contract of a specific type. Because we need type information available to parse data
        # from the read stream, by the time data ends up in the ACS, the lookup will definitely
        # know of the template name.
        names = []  # type: Collection[TypeConName]
        fail_count = 0
        while not names:
            names = self.lookup.template_names(template_name)
            if not names:
                fail_count += 1
                await sleep(1)

                if fail_count % 15 == 0:
                    LOG.warning(
                        "ACS read_async(%r) is still waiting for a package that contains that "
                        "template. If this is taking longer than you expect, it might be because "
                        "the relevant package is not loaded on your ledger.",
                        template_name,
                    )

        unfiltered = self._get_template_state(template_name)
        if len(unfiltered) > 1:
            warnings.warn(
                "Wildcard searches are not supported on async ACS queries",
                UnknownTemplateWarning,
                stacklevel=3,
            )
            unfiltered = dict()

        ((tt, tcd),) = unfiltered.items()

        query = PendingQuery(self.invoker, match, min_count)
        # if the current state is already a match, then don't remember the query since we're
        # already done
        if tcd is None or not query.check_ready(tcd):
            tcd.register_query(query)

        return await await_then(
            query.future, lambda cxds: {cxd.cid: cxd.cdata for cxd in cxds if cxd.cdata is not None}
        )

    def _get_template_state(self, template_name: str) -> "Dict[TypeConName, TemplateContractData]":
        names = self.lookup.template_names(template_name)

        if not names:
            # the warning about unknown template names is really only relevant for wildcard
            # searches; if there is a real package ID, this warning isn't particularly useful
            con = parse_type_con_name(template_name)
            if package_ref(con) == "*":
                warnings.warn(
                    f"Unknown template name: {template_name}", UnknownTemplateWarning, stacklevel=4
                )

        return {name: self._tcdata[name] for name in names}


class TemplateContractData:
    """
    Storage for state related to a specific :class:`Template`.
    """

    def __init__(self):
        # raw ContractId to ContractContextualData
        self._data = dict()  # type: Dict[str, ContractContextualData]
        # outstanding ACS queries
        self._queries = list()  # type: List[PendingQuery]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._data[event.cid.value] = ContractContextualData(
            cid=event.cid, cdata=event.cdata, effective_at=event.time, archived_at=None, active=True
        )

        self._queries = [q for q in self._queries if not q.check_ready(self)]

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        existing_data = self._data.get(event.cid.value)
        effective_at = existing_data.effective_at if existing_data is not None else None
        self._data[event.cid.value] = ContractContextualData(
            cid=event.cid,
            cdata=None,
            effective_at=effective_at,
            archived_at=event.time,
            active=False,
        )

    def get(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        if isinstance(cid, ContractId):
            cid = cid.value
        return self._data.get(cid)

    def subset(
        self, match: ContractMatch, include_archived: bool
    ) -> Collection[ContractContextualData]:
        return [
            cxd
            for cxd in self._data.values()
            if (include_archived or cxd.active)
            and cxd.cdata is not None
            and is_match(match, cxd.cdata)
        ]

    def register_query(self, query: "PendingQuery") -> None:
        self._queries.append(query)


class PendingQuery:
    future: Awaitable[Collection[ContractContextualData]]

    def __init__(self, invoker: Optional[Invoker], match, min_count: int):
        self.future = (
            invoker.create_future() if invoker is not None else get_event_loop().create_future()
        )
        self.match = match
        self.min_count = min_count

    def check_ready(self, tcd: "TemplateContractData") -> bool:
        matches = tcd.subset(self.match, False)
        if len(matches) >= self.min_count:
            cast(Future, self.future).set_result(matches)
            return True
        else:
            return False



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from logging import Logger
from os import PathLike, fspath
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    # We refer to the Config class in a docstring and
    # without this import, Sphinx can't resolve the reference
    # noinspection PyUnresolvedReferences
    from . import Config


class SSLConfig:
    """
    Configuration parameters that affect SSL connections.

    See :meth:`Config.create` for a more detailed description of these parameters.
    """

    def __init__(
        self,
        ca: "Optional[bytes]" = None,
        ca_file: "Optional[PathLike]" = None,
        cert: "Optional[bytes]" = None,
        cert_file: "Optional[PathLike]" = None,
        cert_key: "Optional[bytes]" = None,
        cert_key_file: "Optional[PathLike]" = None,
        logger: Optional[Logger] = None,
    ):
        self._ca: Optional[bytes]
        self._cert: Optional[bytes]
        self._cert_key: Optional[bytes]

        if ca_file:
            if ca:
                raise ValueError("ca and ca_file cannot both be specified at the same time")
            with open(fspath(ca_file), "rb") as f:
                self._ca = f.read()
        else:
            self._ca = ca

        if cert_file:
            if cert:
                raise ValueError("cert and cert_file cannot both be specified at the same time")
            with open(fspath(cert_file), "rb") as f:
                self._cert = f.read()
        else:
            self._cert = cert

        if cert_key_file:
            if cert_key:
                raise ValueError(
                    "cert_key and cert_key_file cannot both be specified at the same time"
                )
            with open(fspath(cert_key_file), "rb") as f:
                self._cert_key = f.read()
        else:
            self._cert_key = cert_key

    def __bool__(self):
        """
        True if SSL settings are supplied; otherwise False if no SSL settings of any kind were
        supplied.
        """
        return bool(self._ca or self._cert or self._cert_key)

    @property
    def ca(self) -> "Optional[bytes]":
        """
        Server certificate authority file.
        """
        return self._ca

    @property
    def cert(self) -> "Optional[bytes]":
        """
        Client certificate file.
        """
        return self._cert

    @property
    def cert_key(self) -> "Optional[bytes]":
        """
        Client certificate and key file.
        """
        return self._cert_key



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import List, Optional, Sequence

from . import daml_lf_1 as lf
from .._gen.com.daml.daml_lf_1_14 import daml_lf_1_pb2 as pblf

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

__all__ = ["ProtobufParser"]

# TODO: Figure out a way to define these literals in daml_lf_1_pb2.pyi where they belong
# fmt: off
PrimCon = Literal[0, 1, 2]
PrimType = Literal[0, 1, 2, 3, 5, 6, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]
PrimLit_RoundingMode = Literal[0, 1, 2, 3, 4, 5, 6, 7]
BuiltinFunction = Literal[0, 1, 2, 3, 6, 107, 108, 109, 110, 111, 121, 122, 7, 8, 9, 10, 11, 12, 20, 21, 96, 97, 98, 99, 100, 101, 124, 125, 126, 127, 128, 129, 130, 23, 24, 25, 147, 33, 34, 112, 36, 37, 67, 89, 39, 40, 113, 42, 43, 68, 90, 45, 46, 114, 48, 49, 69, 91, 51, 52, 115, 54, 55, 70, 92, 57, 58, 116, 60, 61, 71, 63, 94, 95, 103, 104, 117, 136, 93, 72, 73, 74, 75, 76, 77, 118, 119, 78, 79, 80, 120, 81, 82, 83, 84, 85, 86, 87, 123, 131, 132, 133, 134, 135, 88, 102, 105, 106, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146]

# fmt: on


# noinspection PyPep8Naming,PyMethodMayBeStatic
class ProtobufParser:
    def __init__(self, current_package: lf.PackageRef) -> None:
        self.current_package = current_package
        self.current_module = None  # type: Optional[lf.ModuleRef]
        self.interned_strings = []  # type: List[str]
        self.interned_dotted_names = []  # type: List[Sequence[str]]
        self.interned_types = []  # type: List[lf.Type]

    # noinspection PyUnusedLocal
    def parse_Unit(self, pb: pblf.Unit) -> lf.Unit:
        return lf.UNIT

    def parse_ModuleRef(self, pb: pblf.ModuleRef) -> Optional[lf.ModuleRef]:
        sum_name = pb.package_ref.WhichOneof("Sum")
        if sum_name is None:
            return None

        module_name = self._resolve_dotted_name(pb.module_name_dname, pb.module_name_interned_dname)
        if sum_name == "self":
            return lf.ModuleRef(self.current_package, module_name)
        elif sum_name == "package_id_str":
            return lf.ModuleRef(lf.PackageRef(pb.package_ref.package_id_str), module_name)
        elif sum_name == "package_id_interned_str":
            return lf.ModuleRef(
                lf.PackageRef(self.interned_strings[pb.package_ref.package_id_interned_str]),
                module_name,
            )
        else:
            raise ValueError(f"unknown sum type value: {sum_name!r}")

    def parse_TypeConName(self, pb: pblf.TypeConName) -> lf.TypeConName:
        module_ref = self.parse_ModuleRef(pb.module)
        if module_ref is None:
            raise ValueError("missing a valid ModuleRef in a TypeConName definition")
        return lf.TypeConName(
            module_ref,
            self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname).segments,
        )

    def parse_TypeSynName(self, pb: pblf.TypeSynName) -> lf.TypeSynName:
        module_ref = self.parse_ModuleRef(pb.module)
        if module_ref is None:
            raise ValueError("missing a valid ModuleRef in a TypeSynName definition")
        return lf.TypeSynName(
            module_ref,
            self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname).segments,
        )

    def parse_DottedName(self, pb: pblf.DottedName) -> lf.DottedName:
        return lf.DottedName(segments=self._resolve_string_seq(pb.segments, None))

    def parse_ValName(self, pb: pblf.ValName) -> lf.ValName:
        module_ref = self.parse_ModuleRef(pb.module)
        if module_ref is None:
            raise ValueError("missing a valid ModuleRef in a ValName definition")
        return lf.ValName(
            module_ref,
            self._resolve_string_seq(pb.name_dname, pb.name_interned_dname),
        )

    def parse_FieldWithType(self, pb: pblf.FieldWithType) -> lf.FieldWithType:
        """A field definition in a record or a variant associated with a type."""
        return lf.FieldWithType(
            self._resolve_string(pb.field_str, pb.field_interned_str), self.parse_Type(pb.type)
        )

    def parse_VarWithType(self, pb: pblf.VarWithType) -> lf.VarWithType:
        """Binder associated with a type."""
        return lf.VarWithType(
            self._resolve_string(pb.var_str, pb.var_interned_str), self.parse_Type(pb.type)
        )

    def parse_TypeVarWithKind(self, pb: pblf.TypeVarWithKind) -> lf.TypeVarWithKind:
        return lf.TypeVarWithKind(
            self._resolve_string(pb.var_str, pb.var_interned_str), self.parse_Kind(pb.kind)
        )

    def parse_FieldWithExpr(self, pb: pblf.FieldWithExpr) -> lf.FieldWithExpr:
        return lf.FieldWithExpr(
            self._resolve_string(pb.field_str, pb.field_interned_str), self.parse_Expr(pb.expr)
        )

    def parse_Binding(self, pb: pblf.Binding) -> lf.Binding:
        return lf.Binding(self.parse_VarWithType(pb.binder), self.parse_Expr(pb.bound))

    def parse_Kind(self, pb: pblf.Kind) -> lf.Kind:
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "star":
            return lf.Kind(star=self.parse_Unit(pb.star))
        elif sum_name == "arrow":
            return lf.Kind(arrow=self.parse_Kind_Arrow(pb.arrow))
        elif sum_name == "nat":
            return lf.Kind(nat=self.parse_Unit(pb.nat))
        else:
            raise ValueError(f"unknown sum type value: {sum_name!r}")

    def parse_Kind_Arrow(self, pb: pblf.Kind.Arrow) -> lf.Kind.Arrow:
        return lf.Kind.Arrow(
            tuple(self.parse_Kind(param) for param in pb.params), self.parse_Kind(pb.result)
        )

    def parse_PrimType(self, pb: PrimType) -> lf.PrimType:
        return lf.PrimType(pb)

    def parse_Type(self, pb: pblf.Type) -> lf.Type:
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "var":
            return self.parse_Type_Var(pb.var)
        elif sum_name == "con":
            return self.parse_Type_Con(pb.con)
        elif sum_name == "prim":
            return lf.Type(prim=self.parse_Type_Prim(pb.prim))
        elif sum_name == "fun":
            return self.parse_Type_Fun(pb.fun)
        elif sum_name == "forall":
            return lf.Type(forall=self.parse_Type_Forall(pb.forall))
        elif sum_name == "struct":
            return lf.Type(struct=self.parse_Type_Struct(pb.struct))
        elif sum_name == "nat":
            return lf.Type(nat=pb.nat)
        elif sum_name == "syn":
            return lf.Type(syn=self.parse_Type_Syn(pb.syn))
        elif sum_name == "interned":
            return self.interned_types[pb.interned]
        else:
            raise ValueError(f"unknown sum type value: {sum_name!r}")

    def parse_Type_Var(self, pb: pblf.Type.Var) -> lf.Type:
        return lf.Type(
            var=lf.Type.Var(
                self._resolve_string(pb.var_str, pb.var_interned_str),
                tuple(self.parse_Type(arg) for arg in pb.args),
            )
        )

    def parse_Type_Con(self, pb: pblf.Type.Con) -> lf.Type:
        """
        Create a :class:`Type` instance (but may produce something slightly different than the AST
        due to ``Map``/``Optional`` type rewriting).
        """
        tycon = self.parse_TypeConName(pb.tycon)
        args = tuple(self.parse_Type(arg) for arg in pb.args)
        return lf.Type(con=lf.Type.Con(tycon, args))

    def parse_Type_Prim(self, pb: pblf.Type.Prim) -> lf.Type.Prim:
        return lf.Type.Prim(
            self.parse_PrimType(pb.prim), tuple(self.parse_Type(arg) for arg in pb.args)
        )

    def parse_Type_Fun(self, pb: pblf.Type.Fun) -> lf.Type:
        """``fun`` has been deprecated and this code path is now replaced with PrimType.ARROW"""
        last_type = self.parse_Type(pb.result)
        for param in reversed(pb.params):
            last_type = lf.Type(
                prim=lf.Type.Prim(lf.PrimType.ARROW, (self.parse_Type(param), last_type))
            )
        return last_type

    def parse_Type_Forall(self, pb: pblf.Type.Forall) -> lf.Type.Forall:
        return lf.Type.Forall(
            tuple(self.parse_TypeVarWithKind(var) for var in pb.vars), self.parse_Type(pb.body)
        )

    def parse_Type_Struct(self, pb: pblf.Type.Struct) -> lf.Type.Struct:
        return lf.Type.Struct(tuple(self.parse_FieldWithType(field) for field in pb.fields))

    def parse_Type_Syn(self, pb: pblf.Type.Syn) -> lf.Type.Syn:
        return lf.Type.Syn(
            tysyn=self.parse_TypeSynName(pb.tysyn),
            args=tuple(self.parse_Type(arg) for arg in pb.args),
        )

    def parse_PrimCon(self, pb: PrimCon) -> lf.PrimCon:
        if pb == 0:
            return lf.PrimCon.CON_UNIT
        elif pb == 1:
            return lf.PrimCon.CON_FALSE
        elif pb == 2:
            return lf.PrimCon.CON_TRUE
        else:
            raise ValueError(f"unknown enum value: {pb!r}")

    def parse_BuiltinFunction(self, pb: BuiltinFunction) -> lf.BuiltinFunction:
        return lf.BuiltinFunction(pb)

    def parse_PrimLit(self, pb: pblf.PrimLit) -> lf.PrimLit:
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "int64":
            return lf.PrimLit(int64=pb.int64)
        elif sum_name == "decimal_str":
            return lf.PrimLit(decimal=pb.decimal_str)
        elif sum_name == "numeric_interned_str":
            return lf.PrimLit(numeric=self.interned_strings[pb.numeric_interned_str])
        elif sum_name == "text_str":
            return lf.PrimLit(text=pb.text_str)
        elif sum_name == "text_interned_str":
            return lf.PrimLit(text=self.interned_strings[pb.text_interned_str])
        elif sum_name == "timestamp":
            return lf.PrimLit(timestamp=pb.timestamp)
        elif sum_name == "party_str":
            return lf.PrimLit(party=pb.party_str)
        elif sum_name == "party_interned_str":
            return lf.PrimLit(party=self.interned_strings[pb.party_interned_str])
        elif sum_name == "date":
            return lf.PrimLit(date=pb.date)
        elif sum_name == "rounding_mode":
            return lf.PrimLit(rounding_mode=self.parse_PrimLit_RoundingMode(pb.rounding_mode))
        else:
            raise ValueError(f"unknown Sum value: {pb!r}")

    def parse_PrimLit_RoundingMode(self, pb: PrimLit_RoundingMode) -> lf.PrimLit.RoundingMode:
        return lf.PrimLit.RoundingMode(pb)

    def parse_Expr(self, pb: pblf.Expr) -> lf.Expr:
        location = self.parse_Location(pb.location) if pb.HasField("location") else None

        sum_name = pb.WhichOneof("Sum")
        if sum_name == "var_str":
            return lf.Expr(var=pb.var_str, location=location)
        elif sum_name == "var_interned_str":
            return lf.Expr(var=self.interned_strings[pb.var_interned_str], location=location)
        elif sum_name == "val":
            return lf.Expr(val=self.parse_ValName(pb.val), location=location)
        elif sum_name == "builtin":
            return lf.Expr(builtin=self.parse_BuiltinFunction(pb.builtin), location=location)
        elif sum_name == "prim_con":
            return lf.Expr(prim_con=self.parse_PrimCon(pb.prim_con), location=location)
        elif sum_name == "prim_lit":
            return lf.Expr(prim_lit=self.parse_PrimLit(pb.prim_lit), location=location)
        elif sum_name == "rec_con":
            return lf.Expr(rec_con=self.parse_Expr_RecCon(pb.rec_con), location=location)
        elif sum_name == "rec_proj":
            return lf.Expr(rec_proj=self.parse_Expr_RecProj(pb.rec_proj), location=location)
        elif sum_name == "rec_upd":
            return lf.Expr(rec_upd=self.parse_Expr_RecUpd(pb.rec_upd), location=location)
        elif sum_name == "variant_con":
            return lf.Expr(
                variant_con=self.parse_Expr_VariantCon(pb.variant_con), location=location
            )
        elif sum_name == "enum_con":
            return lf.Expr(enum_con=self.parse_Expr_EnumCon(pb.enum_con), location=location)
        elif sum_name == "struct_con":
            return lf.Expr(struct_con=self.parse_Expr_StructCon(pb.struct_con), location=location)
        elif sum_name == "struct_proj":
            return lf.Expr(
                struct_proj=self.parse_Expr_StructProj(pb.struct_proj), location=location
            )
        elif sum_name == "struct_upd":
            return lf.Expr(struct_upd=self.parse_Expr_StructUpd(pb.struct_upd), location=location)
        elif sum_name == "app":
            return lf.Expr(app=self.parse_Expr_App(pb.app), location=location)
        elif sum_name == "ty_app":
            return lf.Expr(ty_app=self.parse_Expr_TyApp(pb.ty_app), location=location)
        elif sum_name == "abs":
            return lf.Expr(abs=self.parse_Expr_Abs(pb.abs), location=location)
        elif sum_name == "ty_abs":
            return lf.Expr(ty_abs=self.parse_Expr_TyAbs(pb.ty_abs), location=location)
        elif sum_name == "case":
            return lf.Expr(case=self.parse_Case(pb.case), location=location)
        elif sum_name == "let":
            return lf.Expr(let=self.parse_Block(pb.let), location=location)
        elif sum_name == "nil":
            return lf.Expr(nil=self.parse_Expr_Nil(pb.nil), location=location)
        elif sum_name == "cons":
            return lf.Expr(cons=self.parse_Expr_Cons(pb.cons), location=location)
        elif sum_name == "update":
            return lf.Expr(update=self.parse_Update(pb.update), location=location)
        elif sum_name == "scenario":
            return lf.Expr(scenario=self.parse_Scenario(pb.scenario), location=location)
        elif sum_name == "optional_none":
            return lf.Expr(
                optional_none=self.parse_Expr_OptionalNone(pb.optional_none), location=location
            )
        elif sum_name == "optional_some":
            return lf.Expr(
                optional_some=self.parse_Expr_OptionalSome(pb.optional_some), location=location
            )
        elif sum_name == "to_any":
            return lf.Expr(to_any=self.parse_Expr_ToAny(pb.to_any), location=location)
        elif sum_name == "from_any":
            return lf.Expr(from_any=self.parse_Expr_FromAny(pb.from_any), location=location)
        elif sum_name == "type_rep":
            return lf.Expr(type_rep=self.parse_Type(pb.type_rep), location=location)
        elif sum_name == "to_any_exception":
            return lf.Expr(
                to_any_exception=self.parse_Expr_ToAnyException(pb.to_any_exception),
                location=location,
            )
        elif sum_name == "from_any_exception":
            return lf.Expr(
                from_any_exception=self.parse_Expr_FromAnyException(pb.from_any_exception),
                location=location,
            )
        elif sum_name == "throw":
            return lf.Expr(throw=self.parse_Expr_Throw(pb.throw), location=location)
        elif sum_name == "experimental":
            return lf.Expr(
                experimental=self.parse_Expr_Experimental(pb.experimental), location=location
            )
        else:
            raise ValueError(f"Unknown type of Expr: {sum_name!r}")

    def parse_Expr_RecCon(self, pb: pblf.Expr.RecCon) -> lf.Expr.RecCon:
        return lf.Expr.RecCon(
            self.parse_Type_Con(pb.tycon).con,
            tuple(self.parse_FieldWithExpr(field) for field in pb.fields),
        )  # length > 0

    def parse_Expr_RecProj(self, pb: pblf.Expr.RecProj) -> lf.Expr.RecProj:
        return lf.Expr.RecProj(
            self.parse_Type_Con(pb.tycon).con,  # Always fully applied
            self._resolve_string(pb.field_str, pb.field_interned_str),
            self.parse_Expr(pb.record),
        )

    def parse_Expr_RecUpd(self, pb: pblf.Expr.RecUpd) -> lf.Expr.RecUpd:
        """Set ``field`` in ``record`` to ``update``."""
        return lf.Expr.RecUpd(
            self.parse_Type_Con(pb.tycon).con,
            self._resolve_string(pb.field_str, pb.field_interned_str),
            self.parse_Expr(pb.record),
            self.parse_Expr(pb.update),
        )

    def parse_Expr_VariantCon(self, pb: pblf.Expr.VariantCon) -> lf.Expr.VariantCon:
        return lf.Expr.VariantCon(
            self.parse_Type_Con(pb.tycon).con,  # Always fully applied
            self._resolve_string(pb.variant_con_str, pb.variant_con_interned_str),
            self.parse_Expr(pb.variant_arg),
        )

    def parse_Expr_StructCon(self, pb: pblf.Expr.StructCon) -> lf.Expr.StructCon:
        return lf.Expr.StructCon(
            tuple(self.parse_FieldWithExpr(field) for field in pb.fields)
        )  # length > 0

    def parse_Expr_EnumCon(self, pb: pblf.Expr.EnumCon) -> lf.Expr.EnumCon:
        return lf.Expr.EnumCon(
            self.parse_TypeConName(pb.tycon),
            self._resolve_string(pb.enum_con_str, pb.enum_con_interned_str),
        )

    def parse_Expr_StructProj(self, pb: pblf.Expr.StructProj) -> lf.Expr.StructProj:
        return lf.Expr.StructProj(
            self._resolve_string(pb.field_str, pb.field_interned_str), self.parse_Expr(pb.struct)
        )

    def parse_Expr_StructUpd(self, pb: pblf.Expr.StructUpd) -> lf.Expr.StructUpd:
        """Set ``field`` in ``tuple`` to ``update``."""
        return lf.Expr.StructUpd(
            self._resolve_string(pb.field_str, pb.field_interned_str),
            self.parse_Expr(pb.struct),
            self.parse_Expr(pb.update),
        )

    def parse_Expr_App(self, pb: pblf.Expr.App) -> lf.Expr.App:
        return lf.Expr.App(
            self.parse_Expr(pb.fun), tuple(self.parse_Expr(arg) for arg in pb.args)
        )  # length > 0

    def parse_Expr_TyApp(self, pb: pblf.Expr.TyApp) -> lf.Expr.TyApp:
        return lf.Expr.TyApp(
            self.parse_Expr(pb.expr), tuple(self.parse_Type(type) for type in pb.types)
        )  # length > 0

    def parse_Expr_Abs(self, pb: pblf.Expr.Abs) -> lf.Expr.Abs:
        return lf.Expr.Abs(
            tuple(self.parse_VarWithType(param) for param in pb.param),  # length > 0
            self.parse_Expr(pb.body),
        )

    def parse_Expr_TyAbs(self, pb: pblf.Expr.TyAbs) -> lf.Expr.TyAbs:
        return lf.Expr.TyAbs(
            tuple(self.parse_TypeVarWithKind(param) for param in pb.param),  # length > 0
            self.parse_Expr(pb.body),
        )

    def parse_Expr_Nil(self, pb: pblf.Expr.Nil) -> lf.Expr.Nil:
        return lf.Expr.Nil(self.parse_Type(pb.type))

    def parse_Expr_Cons(self, pb: pblf.Expr.Cons) -> lf.Expr.Cons:
        return lf.Expr.Cons(
            self.parse_Type(pb.type),
            tuple(self.parse_Expr(front) for front in pb.front),  # length > 0
            self.parse_Expr(pb.tail),
        )

    def parse_Expr_OptionalNone(self, pb: pblf.Expr.OptionalNone) -> lf.Expr.OptionalNone:
        return lf.Expr.OptionalNone(self.parse_Type(pb.type))

    def parse_Expr_OptionalSome(self, pb: pblf.Expr.OptionalSome) -> lf.Expr.OptionalSome:
        return lf.Expr.OptionalSome(self.parse_Type(pb.type), self.parse_Expr(pb.body))

    def parse_Expr_ToAny(self, pb: pblf.Expr.ToAny) -> lf.Expr.ToAny:
        return lf.Expr.ToAny(self.parse_Type(pb.type), self.parse_Expr(pb.expr))

    def parse_Expr_FromAny(self, pb: pblf.Expr.FromAny) -> lf.Expr.FromAny:
        return lf.Expr.FromAny(self.parse_Type(pb.type), self.parse_Expr(pb.expr))

    def parse_Expr_ToAnyException(self, pb: pblf.Expr.ToAnyException) -> lf.Expr.ToAnyException:
        return lf.Expr.ToAnyException(self.parse_Type(pb.type), self.parse_Expr(pb.expr))

    def parse_Expr_FromAnyException(
        self, pb: pblf.Expr.FromAnyException
    ) -> lf.Expr.FromAnyException:
        return lf.Expr.FromAnyException(self.parse_Type(pb.type), self.parse_Expr(pb.expr))

    def parse_Expr_Throw(self, pb: pblf.Expr.Throw) -> lf.Expr.Throw:
        return lf.Expr.Throw(
            return_type=self.parse_Type(pb.return_type),
            exception_type=self.parse_Type(pb.exception_type),
            exception_expr=self.parse_Expr(pb.exception_expr),
        )

    def parse_Expr_Experimental(self, pb: pblf.Expr.Experimental) -> lf.Expr.Experimental:
        return lf.Expr.Experimental(name=pb.name, type=self.parse_Type(pb.type))

    def parse_CaseAlt(self, pb: pblf.CaseAlt) -> lf.CaseAlt:
        body = self.parse_Expr(pb.body)
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "default":
            return lf.CaseAlt(default=self.parse_Unit(pb.default), body=body)
        elif sum_name == "variant":
            return lf.CaseAlt(variant=self.parse_CaseAlt_Variant(pb.variant), body=body)
        elif sum_name == "prim_con":
            return lf.CaseAlt(prim_con=self.parse_PrimCon(pb.prim_con), body=body)
        elif sum_name == "nil":
            return lf.CaseAlt(nil=self.parse_Unit(pb.nil), body=body)
        elif sum_name == "cons":
            return lf.CaseAlt(cons=self.parse_CaseAlt_Cons(pb.cons), body=body)
        elif sum_name == "optional_none":
            return lf.CaseAlt(optional_none=self.parse_Unit(pb.optional_none), body=body)
        elif sum_name == "optional_some":
            return lf.CaseAlt(
                optional_some=self.parse_CaseAlt_OptionalSome(pb.optional_some), body=body
            )
        elif sum_name == "enum":
            return lf.CaseAlt(enum=self.parse_CaseAlt_Enum(pb.enum), body=body)
        else:
            raise ValueError(f"unknown Sum value: {sum_name!r}")

    def parse_CaseAlt_Variant(self, pb: pblf.CaseAlt.Variant) -> lf.CaseAlt.Variant:
        return lf.CaseAlt.Variant(
            self.parse_TypeConName(pb.con),
            self._resolve_string(pb.variant_str, pb.variant_interned_str),
            self._resolve_string(pb.binder_str, pb.binder_interned_str),
        )

    def parse_CaseAlt_Enum(self, pb: pblf.CaseAlt.Enum) -> lf.CaseAlt.Enum:
        return lf.CaseAlt.Enum(
            self.parse_TypeConName(pb.con),
            self._resolve_string(pb.constructor_str, pb.constructor_interned_str),
        )

    def parse_CaseAlt_Cons(self, pb: pblf.CaseAlt.Cons) -> lf.CaseAlt.Cons:
        return lf.CaseAlt.Cons(
            self._resolve_string(pb.var_head_str, pb.var_head_interned_str),
            self._resolve_string(pb.var_tail_str, pb.var_tail_interned_str),
        )

    def parse_CaseAlt_OptionalSome(self, pb: pblf.CaseAlt.OptionalSome) -> lf.CaseAlt.OptionalSome:
        return lf.CaseAlt.OptionalSome(
            self._resolve_string(pb.var_body_str, pb.var_body_interned_str)
        )

    def parse_Case(self, pb: pblf.Case) -> lf.Case:
        return lf.Case(self.parse_Expr(pb.scrut), tuple(self.parse_CaseAlt(alt) for alt in pb.alts))

    def parse_Block(self, pb: pblf.Block) -> lf.Block:
        return lf.Block(
            tuple(self.parse_Binding(binding) for binding in pb.bindings), self.parse_Expr(pb.body)
        )

    def parse_Pure(self, pb: pblf.Pure) -> lf.Pure:
        return lf.Pure(type=self.parse_Type(pb.type), expr=self.parse_Expr(pb.expr))

    def parse_Update(self, pb: pblf.Update) -> lf.Update:
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "pure":
            return lf.Update(pure=self.parse_Pure(pb.pure))
        elif sum_name == "block":
            return lf.Update(block=self.parse_Block(pb.block))
        elif sum_name == "create":
            return lf.Update(create=self.parse_Update_Create(pb.create))
        elif sum_name == "exercise":
            return lf.Update(exercise=self.parse_Update_Exercise(pb.exercise))
        elif sum_name == "exercise_by_key":
            return lf.Update(exercise_by_key=self.parse_Update_ExerciseByKey(pb.exercise_by_key))
        elif sum_name == "fetch":
            return lf.Update(fetch=self.parse_Update_Fetch(pb.fetch))
        elif sum_name == "get_time":
            return lf.Update(get_time=self.parse_Unit(pb.get_time))
        elif sum_name == "lookup_by_key":
            return lf.Update(lookup_by_key=self.parse_Update_RetrieveByKey(pb.lookup_by_key))
        elif sum_name == "fetch_by_key":
            return lf.Update(fetch_by_key=self.parse_Update_RetrieveByKey(pb.fetch_by_key))
        elif sum_name == "embed_expr":
            return lf.Update(embed_expr=self.parse_Update_EmbedExpr(pb.embed_expr))
        elif sum_name == "try_catch":
            return lf.Update(try_catch=self.parse_Update_TryCatch(pb.try_catch))
        else:
            raise ValueError(f"unknown Sum value: {sum_name!r}")

    def parse_Update_Create(self, pb: pblf.Update.Create) -> lf.Update.Create:
        return lf.Update.Create(
            template=self.parse_TypeConName(pb.template), expr=self.parse_Expr(pb.expr)
        )

    def parse_Update_Exercise(self, pb: pblf.Update.Exercise) -> lf.Update.Exercise:
        return lf.Update.Exercise(
            template=self.parse_TypeConName(pb.template),
            choice=self._resolve_string(pb.choice_str, pb.choice_interned_str),
            cid=self.parse_Expr(pb.cid),
            arg=self.parse_Expr(pb.arg),
        )

    def parse_Update_ExerciseByKey(self, pb: pblf.Update.ExerciseByKey) -> lf.Update.ExerciseByKey:
        return lf.Update.ExerciseByKey(
            template=self.parse_TypeConName(pb.template),
            choice=self.interned_strings[pb.choice_interned_str],
            key=self.parse_Expr(pb.key),
            arg=self.parse_Expr(pb.arg),
        )

    def parse_Update_Fetch(self, pb: pblf.Update.Fetch) -> lf.Update.Fetch:
        return lf.Update.Fetch(
            template=self.parse_TypeConName(pb.template), cid=self.parse_Expr(pb.cid)
        )

    def parse_Update_EmbedExpr(self, pb: pblf.Update.EmbedExpr) -> lf.Update.EmbedExpr:
        return lf.Update.EmbedExpr(type=self.parse_Type(pb.type), body=self.parse_Expr(pb.body))

    def parse_Update_RetrieveByKey(self, pb: pblf.Update.RetrieveByKey) -> lf.Update.RetrieveByKey:
        return lf.Update.RetrieveByKey(
            template=self.parse_TypeConName(pb.template), key=self.parse_Expr(pb.key)
        )

    def parse_Update_TryCatch(self, pb: pblf.Update.TryCatch) -> lf.Update.TryCatch:
        return lf.Update.TryCatch(
            return_type=self.parse_Type(pb.return_type),
            try_expr=self.parse_Expr(pb.try_expr),
            var=self.interned_strings[pb.var_interned_str],
            catch_expr=self.parse_Expr(pb.catch_expr),
        )

    def parse_Scenario(self, pb: pblf.Scenario) -> lf.Scenario:
        sum_name = pb.WhichOneof("Sum")
        if sum_name == "pure":
            return lf.Scenario(pure=self.parse_Pure(pb.pure))
        elif sum_name == "block":
            return lf.Scenario(block=self.parse_Block(pb.block))
        elif sum_name == "commit":
            return lf.Scenario(commit=self.parse_Scenario_Commit(pb.commit))
        elif sum_name == "mustFailAt":
            return lf.Scenario(must_fail_at=self.parse_Scenario_Commit(pb.mustFailAt))
        elif sum_name == "pass":
            return lf.Scenario(pass_=self.parse_Expr(getattr(pb, "pass")))
        elif sum_name == "get_time":
            return lf.Scenario(get_time=self.parse_Unit(pb.get_time))
        elif sum_name == "get_party":
            return lf.Scenario(get_party=self.parse_Expr(pb.get_party))
        elif sum_name == "embed_expr":
            return lf.Scenario(embed_expr=self.parse_Scenario_EmbedExpr(pb.embed_expr))
        else:
            raise ValueError("unknown Sum value")

    def parse_Scenario_Commit(self, pb: pblf.Scenario.Commit) -> lf.Scenario.Commit:
        return lf.Scenario.Commit(
            party=self.parse_Expr(pb.party),
            expr=self.parse_Expr(pb.expr),
            ret_type=self.parse_Type(pb.ret_type),
        )

    def parse_Scenario_EmbedExpr(self, pb: pblf.Scenario.EmbedExpr) -> lf.Scenario.EmbedExpr:
        return lf.Scenario.EmbedExpr(type=self.parse_Type(pb.type), body=self.parse_Expr(pb.body))

    def parse_Location(self, pb: pblf.Location) -> lf.Location:
        module_ref = self.parse_ModuleRef(pb.module)
        if module_ref is None:
            module_ref = self.current_module
        if module_ref is None:
            # This is probably a programming mistake, since nowhere in the Daml-LF protobuf does
            # a Location occur outside of a Module
            raise ValueError(
                "cannot parse a Location object without a ModuleRef outside of a Module"
            )

        return lf.Location(module_ref, self.parse_Location_Range(pb.range))

    def parse_Location_Range(self, pb: pblf.Location.Range) -> lf.Location.Range:
        return lf.Location.Range(pb.start_line, pb.start_col, pb.end_line, pb.end_col)

    def parse_TemplateChoice(self, pb: pblf.TemplateChoice) -> lf.TemplateChoice:
        return lf.TemplateChoice(
            name=self._resolve_string(pb.name_str, pb.name_interned_str),
            consuming=pb.consuming,
            controllers=self.parse_Expr(pb.controllers),
            observers=self.parse_Expr(pb.observers) if pb.HasField("observers") else None,
            arg_binder=self.parse_VarWithType(pb.arg_binder),
            ret_type=self.parse_Type(pb.ret_type),
            update=self.parse_Expr(pb.update),
            self_binder=self._resolve_string(pb.self_binder_str, pb.self_binder_interned_str),
            location=self.parse_Location(pb.location),
        )

    def parse_KeyExpr(self, pb: pblf.KeyExpr) -> lf.KeyExpr:
        if pb.HasField("projections"):
            return lf.KeyExpr(projections=self.parse_KeyExpr_Projections(pb.projections))
        elif pb.HasField("record"):
            return lf.KeyExpr(record=self.parse_KeyExpr_Record(pb.record))
        else:
            raise ValueError(f"unknown KeyExpr {pb}")

    def parse_KeyExpr_Projection(self, pb: pblf.KeyExpr.Projection) -> lf.KeyExpr.Projection:
        return lf.KeyExpr.Projection(
            tycon=self.parse_Type_Con(pb.tycon).con,
            field=self._resolve_string(pb.field_str, pb.field_interned_str),
        )

    def parse_KeyExpr_Projections(self, pb: pblf.KeyExpr.Projections) -> lf.KeyExpr.Projections:
        return lf.KeyExpr.Projections(
            projections=[self.parse_KeyExpr_Projection(p) for p in pb.projections]
        )

    def parse_KeyExpr_RecordField(self, pb: pblf.KeyExpr.RecordField) -> lf.KeyExpr.RecordField:
        return lf.KeyExpr.RecordField(
            field=self._resolve_string(pb.field_str, pb.field_interned_str),
            expr=self.parse_KeyExpr(pb.expr),
        )

    def parse_KeyExpr_Record(self, pb: pblf.KeyExpr.Record) -> lf.KeyExpr.Record:
        return lf.KeyExpr.Record(
            tycon=self.parse_Type_Con(pb.tycon).con,
            fields=[self.parse_KeyExpr_RecordField(p) for p in pb.fields],
        )

    def parse_DefTemplate(self, pb: pblf.DefTemplate) -> lf.DefTemplate:
        return lf.DefTemplate(
            tycon=self._resolve_dotted_name(pb.tycon_dname, pb.tycon_interned_dname),
            param=self._resolve_string(pb.param_str, pb.param_interned_str),
            precond=self.parse_Expr(pb.precond),
            signatories=self.parse_Expr(pb.signatories),
            agreement=self.parse_Expr(pb.agreement),
            choices=tuple(self.parse_TemplateChoice(choice) for choice in pb.choices),
            observers=self.parse_Expr(pb.observers),
            location=self.parse_Location(pb.location),
            key=self.parse_DefTemplate_DefKey(pb.key) if pb.HasField("key") else None,
        )

    def parse_DefTemplate_DefKey(self, pb: pblf.DefTemplate.DefKey) -> lf.DefTemplate.DefKey:
        kwargs = dict(type=self.parse_Type(pb.type), maintainers=self.parse_Expr(pb.maintainers))
        key_expr_name = pb.WhichOneof("key_expr")
        if key_expr_name == "key":
            kwargs["key"] = self.parse_KeyExpr(pb.key)
        elif key_expr_name == "complex_key":
            kwargs["complex_key"] = self.parse_Expr(pb.complex_key)
        return lf.DefTemplate.DefKey(**kwargs)

    def parse_DefDataType(self, pb: pblf.DefDataType) -> lf.DefDataType:
        name = self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname)
        params = tuple(self.parse_TypeVarWithKind(param) for param in pb.params)
        serializable = pb.serializable
        location = self.parse_Location(pb.location)

        DataCons_name = pb.WhichOneof("DataCons")
        if DataCons_name == "record":
            return lf.DefDataType(
                name=name,
                params=params,
                serializable=serializable,
                location=location,
                record=self.parse_DefDataType_Fields(pb.record),
            )
        elif DataCons_name == "variant":
            return lf.DefDataType(
                name=name,
                params=params,
                serializable=serializable,
                location=location,
                variant=self.parse_DefDataType_Fields(pb.variant),
            )
        elif DataCons_name == "enum":
            return lf.DefDataType(
                name=name,
                params=params,
                serializable=serializable,
                location=location,
                enum=self.parse_DefDataType_EnumConstructors(pb.enum),
            )
        else:
            raise ValueError(f"unknown DataCons value: {DataCons_name!r}")

    def parse_DefDataType_Fields(self, pb: pblf.DefDataType.Fields) -> lf.DefDataType.Fields:
        return lf.DefDataType.Fields(
            fields=tuple(self.parse_FieldWithType(field) for field in pb.fields)
        )

    def parse_DefDataType_EnumConstructors(
        self, pb: pblf.DefDataType.EnumConstructors
    ) -> lf.DefDataType.EnumConstructors:
        ctors_1 = tuple(pb.constructors_str)
        ctors_2 = tuple(self.interned_strings[idx] for idx in pb.constructors_interned_str)
        return lf.DefDataType.EnumConstructors(constructors=ctors_1 + ctors_2)

    def parse_DefTypeSyn(self, pb: pblf.DefTypeSyn) -> lf.DefTypeSyn:
        return lf.DefTypeSyn(
            name=self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname),
            params=tuple(self.parse_TypeVarWithKind(param) for param in pb.params),
            type=self.parse_Type(pb.type),
            location=self.parse_Location(pb.location),
        )

    def parse_DefValue(self, pb: pblf.DefValue) -> lf.DefValue:
        return lf.DefValue(
            name_with_type=self.parse_DefValue_NameWithType(pb.name_with_type),
            expr=lambda: self.parse_Expr(pb.expr),
            no_party_literals=pb.no_party_literals,
            is_test=pb.is_test,
            location=self.parse_Location(pb.location),
        )

    def parse_DefValue_NameWithType(
        self, pb: pblf.DefValue.NameWithType
    ) -> lf.DefValue.NameWithType:
        return lf.DefValue.NameWithType(
            name=self._resolve_string_seq(pb.name_dname, pb.name_interned_dname),
            type=self.parse_Type(pb.type),
        )

    def parse_FeatureFlags(self, pb: pblf.FeatureFlags) -> lf.FeatureFlags:
        return lf.FeatureFlags(
            forbid_party_literals=pb.forbidPartyLiterals,
            dont_divulge_contract_ids_in_create_arguments=pb.dontDivulgeContractIdsInCreateArguments,
            dont_disclose_nonconsuming_choices_to_observers=pb.dontDiscloseNonConsumingChoicesToObservers,
        )

    def parse_Module(self, pb: pblf.Module) -> lf.Module:
        name = self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname)
        self.current_module = lf.ModuleRef(self.current_package, name)
        child_parser = self._copy()
        try:
            module = lf.Module(
                name=name,
                flags=child_parser.parse_FeatureFlags(pb.flags),
                synonyms=tuple(child_parser.parse_DefTypeSyn(value) for value in pb.synonyms),
                data_types=tuple(
                    child_parser.parse_DefDataType(data_type) for data_type in pb.data_types
                ),
                values=tuple(child_parser.parse_DefValue(value) for value in pb.values),
                templates=tuple(
                    child_parser.parse_DefTemplate(template) for template in pb.templates
                ),
            )
        finally:
            self.current_module = None

        return module

    def parse_Package(self, pb: pblf.Package) -> lf.Package:
        # TODO: this modifies state in a parser which is less than ideal; a better pattern would be
        #  to create a sub-parser with the contextual state required to understand interned package
        #  IDs
        self.interned_strings.extend(pb.interned_strings)

        indices = [
            tuple(self.interned_strings[idx] for idx in idn.segments_interned_str)
            for idn in pb.interned_dotted_names
        ]

        self.interned_dotted_names.extend(indices)

        # types in the type intern table are allowed to refer to previously interned types, so we
        # must parse, then add each type individually
        for type_pb in pb.interned_types:
            self.interned_types.append(self.parse_Type(type_pb))

        return lf.Package(
            modules=tuple(self.parse_Module(module) for module in pb.modules),
            metadata=self.parse_PackageMetadata(pb.metadata) if pb.HasField("metadata") else None,
        )

    def parse_PackageMetadata(self, pb: pblf.PackageMetadata) -> lf.PackageMetadata:
        return lf.PackageMetadata(
            name=self.interned_strings[pb.name_interned_str],
            version=self.interned_strings[pb.version_interned_str],
        )

    def _copy(self) -> "ProtobufParser":
        p = ProtobufParser(self.current_package)
        p.current_module = self.current_module
        p.interned_strings = self.interned_strings
        p.interned_dotted_names = self.interned_dotted_names
        p.interned_types = self.interned_types
        return p

    def _resolve_string(self, name: Optional[str], interned_id: Optional[int]) -> str:
        # note that we intentionally conflate None and empty string, or None and 0 because
        # of Protobuf
        return name if name else self.interned_strings[interned_id or 0]

    def _resolve_string_seq(
        self, name: Optional[Sequence[str]], name_interned_id: Optional[int]
    ) -> Sequence[str]:
        if self.interned_dotted_names:
            return tuple(name) if name else self.interned_dotted_names[name_interned_id or 0]
        else:
            return tuple(name) if name else tuple()

    def _resolve_dotted_name(
        self, pb_dotted_name: pblf.DottedName, interned_id: Optional[int]
    ) -> lf.DottedName:
        return lf.DottedName(self._resolve_string_seq(pb_dotted_name.segments, interned_id))



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import collections.abc
from typing import Any, Optional, Type as PyType, TypeVar

from google.protobuf import timestamp_pb2
from google.protobuf.empty_pb2 import Empty

from .._gen.com.daml.ledger.api import v1 as lapipb
from ..damlast.daml_lf_1 import DefDataType, Type
from ..damlast.util import find_variant_type
from ..prim import (
    FrozenDict,
    date_to_int,
    datetime_to_epoch_microseconds,
    decimal_to_str,
    to_bool,
    to_date,
    to_datetime,
    to_decimal,
    to_int,
    to_party,
    to_str,
    to_variant,
)
from .context import Context
from .mapper import ValueMapper

__all__ = ["ProtobufDecoder", "ProtobufEncoder", "get_value", "set_value"]

T = TypeVar("T")


class ProtobufDecoder(ValueMapper):
    """
    Convert value_pb2.Value (or any of its cases) to a native Python type.

    This mapper also handles:
     * re-typing ContractIds (they come over the Ledger API without their type parameter)
     * non-verbose Ledger API record values (field names are omitted for message brevity, meaning
       they can only be understood in a generic way by knowing the metadata associated with the
       object ahead of time)
    """

    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        msg = get_value(obj, "record", lapipb.Record)

        d = dict()
        for fld_metadata, fld_data in zip(record.fields, msg.fields):
            name = fld_metadata.field
            value = context.append_path(name).convert(fld_metadata.type, fld_data.value)
            d[name] = value

        return d

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        msg = get_value(obj, "variant", lapipb.Variant)
        obj_ctor = msg.constructor
        obj_value = msg.value
        obj_type = find_variant_type(dt, variant, obj_ctor)

        d_value = context.append_path(obj_ctor).convert(obj_type, obj_value)

        return {obj_ctor: d_value}

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        msg = get_value(obj, "enum", lapipb.Enum)
        return context.value_validate_enum(msg.constructor, enum)

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        return {}

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        return get_value(obj, "bool", bool)

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        return get_value(obj, "int64", int)

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        return get_value(obj, "text", str)

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        return to_datetime(get_value(obj, "timestamp", timestamp_pb2.Timestamp))

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        return to_party(get_value(obj, "party", str))

    def prim_list(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = get_value(obj, "list", lapipb.List)
        return context.convert_list(item_type, msg.elements)

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        msg = get_value(obj, "date", str)
        return to_date(msg)

    def prim_contract_id(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = get_value(obj, "contract_id", str)
        return context.convert_contract_id(item_type, msg)

    def prim_optional(self, context: "Context", t: "Type", obj: "Any") -> "Any":
        msg = get_value(obj, "optional", lapipb.Optional)
        maybe_val = msg.value if msg.HasField("value") else None
        return context.convert_optional(t, maybe_val)

    def prim_text_map(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = get_value(obj, "map", lapipb.Map)
        mapping = {entry_pb.key: entry_pb.value for entry_pb in msg.entries}
        return context.convert_text_map(item_type, mapping)

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        msg = get_value(obj, "numeric", str)
        return to_decimal(msg)

    def prim_gen_map(
        self, context: "Context", key_type: "Type", value_type: "Type", obj: "Any"
    ) -> "Any":
        msg = get_value(obj, "gen_map", lapipb.GenMap)

        obj = {}
        for i, entry in enumerate(msg.entries):
            key = context.append_path(f"[key: {i}]").convert(key_type, entry.key)
            value = context.append_path(f"[{key}]").convert(value_type, entry.value)
            if isinstance(key, collections.abc.Mapping):
                # Python dicts cannot be the keys of a dict because they are not hashable;
                # FrozenDict wraps a dict with a trivial hashing implementation to avoid
                # problems
                obj[FrozenDict(key)] = value
            else:
                obj[key] = value

        return obj


class ProtobufEncoder(ValueMapper):
    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        msg = lapipb.Record()

        for fld in record.fields:
            entry = msg.fields.add()
            entry.label = fld.field
            ctor, val = context.append_path(fld.field).convert(fld.type, obj[fld.field])
            set_value(entry.value, ctor, val)

        return "record", msg

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        obj_ctor, obj_value = to_variant(obj)
        obj_type = find_variant_type(dt, variant, obj_ctor)

        msg_case, msg_value = context.append_path(obj_ctor).convert(obj_type, obj_value)

        msg = lapipb.Variant()
        msg.constructor = obj_ctor
        set_value(msg.value, msg_case, msg_value)

        return "variant", msg

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        msg = lapipb.Enum()
        msg.constructor = context.value_validate_enum(obj, enum)
        return "enum", msg

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        return "unit", Empty()

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        return "bool", to_bool(obj)

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        return "int64", to_int(obj)

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        return "text", to_str(obj)

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        return "timestamp", datetime_to_epoch_microseconds(to_datetime(obj))

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        return "party", to_str(obj)

    def prim_list(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = lapipb.List()
        for i, item in enumerate(obj):
            value = msg.elements.add()
            ctor, val = context.append_path(f"[{i}").convert(item_type, item)
            set_value(value, ctor, val)
        return "list", msg

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        return "date", date_to_int(to_date(obj))

    def prim_contract_id(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return "contract_id", to_str(obj)

    def prim_optional(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = lapipb.Optional()
        if obj is not None:
            ctor, val = context.append_path("?").convert(item_type, obj)
            set_value(msg.value, ctor, val)
        return "optional", msg

    def prim_text_map(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        msg = lapipb.Map()
        for key, value in obj.items():
            entry = msg.entries.add()
            entry.key = key
            ctor, val = context.append_path(f"[{key}]").convert(item_type, value)
            set_value(entry.value, ctor, val)
        return "map", msg

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        d = to_decimal(obj)
        return "numeric", decimal_to_str(d) if d is not None else None

    def prim_gen_map(
        self, context: "Context", key_type: "Type", value_type: "Type", obj: "Any"
    ) -> "Any":
        msg = lapipb.GenMap()
        for i, (key, value) in enumerate(obj.items()):
            entry = msg.entries.add()
            key_ctor, key_val = context.append_path(f"[key: {i}]").convert(key_type, key)
            val_ctor, val_val = context.append_path(f"[{key}]").convert(value_type, value)
            set_value(entry.key, key_ctor, key_val)
            set_value(entry.value, val_ctor, val_val)
        return "gen_map", msg


def get_value(obj: "Any", field: str, pb_type: "PyType[T]") -> "T":
    return obj if isinstance(obj, pb_type) else getattr(obj, field)


def set_value(message: "lapipb.Value", ctor: "Optional[str]", value: "Any") -> None:
    """
    Work around the somewhat crazy API of Python's gRPC library to apply a known value to a
    :class:`Value`.

    :param message:
        The :class:`Value` object to modify.
    :param ctor:
        The actual field to apply to, or ``None`` to interpret the entire message as a ``Record``
        instead.
    :param value:
        The actual value to set. Must be compatible with the appropriate field.
    """
    try:
        if ctor is None:
            message.MergeFrom(value)
        elif ctor == "unit":
            message.unit.SetInParent()
        elif ctor in ("record", "variant", "list", "optional", "enum", "map", "gen_map"):
            getattr(message, ctor).MergeFrom(value)
        else:
            setattr(message, ctor, value)
    except:  # noqa
        from .. import LOG

        LOG.error("Failed to set a value %s, %s", ctor, value)
        raise



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.descriptor import EnumDescriptor
from google.protobuf.message import Message as _Message

from .daml_lf_1_pb2 import Package

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "ArchivePayload",
    "Archive",
]

class HashFunction:
    DESCRIPTOR: _typing.ClassVar[EnumDescriptor] = ...
    SHA256: _typing.ClassVar[_L[0]] = ...
SHA256 = _L[0]


class ArchivePayload(_Message):
    minor: _builtins.str
    @property
    def daml_lf_1(self) -> Package: ...
    @_typing.overload
    def __init__(self, *, minor: _typing.Optional[_builtins.str] = ...): ...
    @_typing.overload
    def __init__(self, *, minor: _typing.Optional[_builtins.str] = ..., daml_lf_1: Package = ...): ...
    def HasField(self, field_name: _L["minor", "Sum", "daml_lf_1"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["minor", "Sum", "daml_lf_1"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["Sum"]) -> _L[None, "daml_lf_1"]: ...

class Archive(_Message):
    @property
    def hash_function(self) -> _L[0]: ...
    payload: _builtins.bytes
    hash: _builtins.str
    def __init__(self, *, hash_function: _typing.Optional[_L['SHA256', 0]] = ..., payload: _typing.Optional[_builtins.bytes] = ..., hash: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["hash_function", "payload", "hash"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["hash_function", "payload", "hash"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Collection, Mapping

from ..client import _NetworkImpl
from ..prim import Party

if TYPE_CHECKING:
    from aiohttp import web


def build_routes(network_impl: "_NetworkImpl", registry=None) -> "Collection[web.AbstractRouteDef]":
    # this import is optional, so do it as late as possible
    from aiohttp import web

    routes = web.RouteTableDef()

    from prometheus_client import REGISTRY, exposition

    if registry is None:
        registry = REGISTRY

    @dataclass(frozen=True)
    class ReadyResult:
        parties: "Mapping[Party, ReadyPartyResult]"

    @dataclass(frozen=True)
    class ReadyPartyResult:
        url: str
        ready: bool

    @routes.get("/metrics")
    async def get_metrics(request: "web.Request") -> "web.Response":
        encoder, content_type = exposition.choose_encoder(request.headers["Accept"])
        metrics_bytes: bytes = encoder(registry)
        return web.HTTPOk(body=metrics_bytes, content_type="text/plain; version=0.0.4")

    @routes.get("/healthz")
    async def get_health(_: "web.Request") -> "web.Response":
        return web.json_response({})

    @routes.get("/readyz")
    async def get_ready(_: "web.Request") -> "web.Response":
        parties = {
            party_impl.party: ReadyPartyResult(
                ready=party_impl.ready().done(), url=party_impl.resolved_config().url
            )
            for party_impl in network_impl.party_impls()
        }
        return web.json_response(asdict(ReadyResult(parties=parties)))

    return routes
